import tensorflow as tf
import numpy as np
import cv2
import os.path
import argparse

parser = argparse.ArgumentParser()
parser.add_argument("--input_path", default="input_example/original.png", help="Path to the image to decompose.")
parser.add_argument("--output_path", default="output_example", help="Directory where you wish the output to reside.")
parser.add_argument("--model_path", default="model", help="Directory with containing a pre-trained model for testing.")
parser.add_argument("--input_gamma", type=float, default=2.2, help="Gamma of the input image.")
parser.add_argument("--size", type=int, default=256, help="Desired width and height of the output. Must be a power of 2.")
p = parser.parse_args()

ROWS = p.size
COLUMNS = p.size
BATCH_SIZE = 1
CHANNELS = 4

def gamma_correction(img, correction):
    return cv2.pow(img, correction)

def load_image(filename):
    image = cv2.imread(filename, cv2.IMREAD_UNCHANGED).astype('float32')
    height, width, channels = image.shape
    if (ROWS != height or COLUMNS != width):
        image = cv2.resize(image, (ROWS, COLUMNS), interpolation=cv2.INTER_AREA)
    if (channels == 3):
        image = np.concatenate((image, np.ones([ROWS, COLUMNS, 1]) * 255.0), axis=2)
    if (channels == 1):
        raise Exception("Greyscale images are not supported.")
    image = np.asarray(image, dtype=np.float32)
    # Trained at gamma 1.6, assumes gamma 2.2 input.
    image = gamma_correction(image / 255.0, p.input_gamma / 1.6)
    return image

def save_image(npdata, filename):
    success = cv2.imwrite(filename, npdata * 255.0)

# 2D convolution wrapper
def conv2d(input_batch, out_channels, kernel, strides):
    in_channels = input_batch.get_shape()[3]
    shape = [kernel, kernel, in_channels, out_channels]
    filter = tf.get_variable("filter", shape, initializer=tf.contrib.layers.xavier_initializer())
    bias = tf.get_variable("bias", out_channels, initializer=tf.contrib.layers.xavier_initializer())
    conv = tf.nn.conv2d(input_batch, filter, strides=[1, strides, strides, 1], padding='SAME')
    return tf.nn.bias_add(conv, bias) 

# Up-convolution wrapper
def upconvolution(input_batch, out_channels, kernel, size):
    upsample = tf.image.resize_images(input_batch, size, method=tf.image.ResizeMethod.NEAREST_NEIGHBOR)
    conv = conv2d(upsample, out_channels, kernel, strides=1)
    return conv

# Build the model
def conv_net(input_batch):
    
    # Used for skip-connections
    encoder_layers = []
    # Encoder
    encoder_n = 8
    encoder_output = input_batch
    for i in range(0, encoder_n):
        with tf.variable_scope("encoder_%d" % i):
            kernel = 3 if i > 1 else 5
            strides = 2 if i > 0 else 1
            # 32, 32, 32, 32, 64, 64, 64, 128
            out_channels = 2 ** ((i - 1) / 3 + 5) if i > 0 else 2 ** 5
            conv_output = conv2d(encoder_output, out_channels, kernel, strides)
            encoder_output = tf.nn.relu(conv_output)
            encoder_layers.append(encoder_output)
    
    # Substitutes for fully connected layers
    flat_n = 2
    flat_output = encoder_output
    for i in range(0, flat_n):
        with tf.variable_scope("flat_%d" % i):
            # 128 * 2, 128 * 4
            out_channels = 2 ** (i + 8)
            flat_output = conv2d(flat_output, out_channels, 3, strides=1)
    
    # Decoder
    decoder_n = 7
    decoder_output = flat_output
    for i in range(0, decoder_n):
        with tf.variable_scope("decoder_%d" % i):
            kernel = 3 if i != decoder_n - 1 else 5
            reverse_i = decoder_n - i - 1
            size = [ROWS / 2 ** (decoder_n - i - 1), COLUMNS / 2 ** (decoder_n - i - 1)]
            # 64 * 4, 64 * 4, 64 * 4, 32 * 4, 32 * 4, 32 * 4, 32 * 4
            out_channels = 2 ** ((decoder_n - i - 2) / 3 + 7) if i != decoder_n - 1 else 2 ** 7
            upconv_output = upconvolution(decoder_output, out_channels, kernel, size)
            relu_output = tf.nn.relu(upconv_output)
        with tf.variable_scope("concat_%d" % i):
            kernel = 3
            # 64 * 4, 64 * 4, 64 * 4, 32 * 5, 32 * 5, 32 * 5, 3 * 4
            out_channels = 2 ** ((decoder_n - i - 2) / 3 + 7) if i != decoder_n - 1 else 2 ** 2 * 3
            out_channels = out_channels + (2 ** 5 if out_channels == 2 ** 7 else 0)
            concat_output = tf.concat(3, [relu_output, encoder_layers[decoder_n - i - 1]])
            conv_output = conv2d(concat_output, out_channels, kernel, strides=1)
            decoder_output = tf.nn.relu(conv_output) if i != decoder_n - 1 else tf.nn.sigmoid(conv_output)
    
    return decoder_output

def main():
    input_image = load_image(p.input_path)
    # Placeholder for the input
    original_image_input = tf.placeholder(tf.float32, [BATCH_SIZE, ROWS, COLUMNS, CHANNELS])
    # Setup the network
    prediction_all = conv_net(original_image_input[:, :, :, 0:3])
    # Split the result into layers
    mask_2d = tf.expand_dims(original_image_input[:, :, :, 3], 3)
    mask = tf.concat(3, [mask_2d, mask_2d, mask_2d])
    prediction_s = tf.concat(3, [tf.add(tf.mul(prediction_all[:, :, :, 0: 3], mask), tf.sub(1.0, mask)), mask_2d])
    prediction_i = tf.concat(3, [tf.add(tf.mul(prediction_all[:, :, :, 3: 6], mask), tf.sub(1.0, mask)), mask_2d])
    prediction_o = tf.concat(3, [tf.add(tf.mul(prediction_all[:, :, :, 6: 9], mask), tf.sub(1.0, mask)), mask_2d])
    prediction_a = tf.concat(3, [tf.add(tf.mul(prediction_all[:, :, :, 9:12], mask), tf.sub(1.0, mask)), mask_2d])

    # Create a saver
    saver = tf.train.Saver(tf.all_variables())
    # Setup session and init variables
    session = tf.Session()
    session.run(tf.initialize_all_variables())
    
    # Create a batch (in this revised test script the batch size is always 1)
    original_image_batch = np.zeros([BATCH_SIZE, ROWS, COLUMNS, CHANNELS])
    original_image_batch[0, :, :, :] = input_image

    # Load a pre-trained model
    ckpt = tf.train.get_checkpoint_state(p.model_path)
    if ckpt and ckpt.model_checkpoint_path:
        saver.restore(session, ckpt.model_checkpoint_path)
    
    # Run the network
    layers_output = []
    layers_output = list(session.run([prediction_i, prediction_o, prediction_a, prediction_s], 
                                feed_dict={original_image_input : original_image_batch}))

    if not os.path.exists(p.output_path):
        os.makedirs(p.output_path)
        
    if not os.path.exists(p.output_path + "/gamma"):
        os.makedirs(p.output_path + "/gamma")
    if not os.path.exists(p.output_path + "/linear"):
        os.makedirs(p.output_path + "/linear")
        
    # HTML and output generation
    f = open(p.output_path + '/viewer' + '.html', 'w')
    html = '<!DOCTYPE html>' + \
           '<html>' + \
               '<head>' + \
                   '<meta charset=utf-8 />' + \
                   '<title></title>' + \
                   '<style>' + \
                       'div.container {display: inline-block; width: 450px; height: 450px; margin: 10px; }' + \
                       'p { text-align:center; }' + \
                   '</style>' +  \
                '</head>' + \
                '<body>'
    
    layer_names = ["irradiance", "ao", "albedo", "specular", "combined"]
    combined_layers = np.ones([BATCH_SIZE, ROWS, COLUMNS, CHANNELS])
    for i in range(0, len(layer_names)):
        if i != len(layer_names) - 1:
            layer = layers_output[i]
            combined_layers = combined_layers * layer if i != len(layer_names) - 2 else combined_layers + layer
            save_image(layer[0, :, :, :], p.output_path + "/linear/" + layer_names[i] + ".png")
        else:
            layer = combined_layers
        gamma_layer = gamma_correction(layer[0, :, :, :], 1.0 / p.input_gamma)
        save_image(gamma_layer, p.output_path + "/gamma/" + layer_names[i] + ".png")
        html += '<p style="float: left; font-size: 12pt; text-align: center;' + \
                          'font-family: Arial, Helvetica, sans-serif; width: 19%;' + \
                          'margin-right: ' + ('0.5' if i != len(layer_names) - 1 else '0') +'%;' + \
                          'margin-bottom: 0.5em;">' + \
                    '<img src="gamma/' + layer_names[i] + '.png" style="width: 100%">' + layer_names[i] + \
                '</p>'
                
    html +=     '<p style="clear: both;">' + \
                '</body>' + \
                '</html>'
    f.write(html)
    f.close()
    
main()